using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Text;
using System.Data;
using System.Drawing;
using System.Xml;
using System.Web;
using System.Windows.Forms;
using System.IO;
using C1.C1Preview;

namespace C1SampleViewer
{
    public class C1SampleDocument : C1PrintDocument
	{
		private string baseFont = "Trebuchet MS";
        private SortedList<string, SortedList> indexList = null;
        private bool firstPage = false;
        private RenderToc toc = null;

        public C1SampleDocument()
        {
        }

        public void LoadFromSettings()
        {
            // Clear existing document contents
            Clear();
            
            // Use the same header for all pages
            PageLayout.PageHeader = CreatePageHeader();

            // Add frequently used images to the dictionary
            Dictionary.Clear();
            Dictionary.Add(new DictionaryImage("CS", C1SampleViewer.Properties.Resources.CS));
            Dictionary.Add(new DictionaryImage("VB", C1SampleViewer.Properties.Resources.VB));
            Dictionary.Add(new DictionaryImage("XML", C1SampleViewer.Properties.Resources.XML));

            // Get settings properties
            bool bLocal = C1SampleViewer.Properties.Settings.Default.Local;
            bool bToc = C1SampleViewer.Properties.Settings.Default.Toc;
            bool bIndex = C1SampleViewer.Properties.Settings.Default.Index;
            string path = C1SampleViewer.Properties.Settings.Default.LocalPath;

            // Create a SortedList for indexing the samples alphabetically, if desired
            indexList = bIndex ? new SortedList<string, SortedList>() : null;

            // Create a table of contents, if desired
            toc = bToc ? CreateToc() : null;

            // Load samples from the local machine or the HelpCentral site, based on settings
            bool ok;
            if (bLocal)
            {
                ok = LoadLocalSamples(path);
                DocumentInfo.Title = "Local Samples";
            }
            else
            {
                ok = LoadRemoteSamples(@"http://helpcentral.componentone.com/samples.xml");
                DocumentInfo.Title = "HelpCentral Samples";
            }

            // Build the index pages, if desired
            if (ok && bIndex)
                CreateIndex();

            // Clear the document if no samples were found
            if (!ok)
            {
                Clear();
                Body.Children.Add(new RenderText("No samples were found."));
            }

            // Generate the document
            Generate();
        }

        public void AddToIndex(string name, string product, RenderObject ro)
        {
            string letter = name[0].ToString().ToUpper();

            if (!indexList.ContainsKey(letter))
                indexList.Add(letter, new SortedList());

            SortedList list = indexList[letter];
            string fmt = string.Format("{0} ({1})", name, product);
            list.Add(fmt, ro);
        }

        private bool LoadLocalSamples(string path)
        {
            bool ok = false;

            try
            {
                DirectoryInfo di = new DirectoryInfo(path);
                DirectoryInfo[] folders = di.GetDirectories();

                for (int i = 0; i < folders.Length; i++)
                {
                    string name = folders[i].Name;
                    string samples = folders[i].FullName + "\\Samples";

                    if (Directory.Exists(samples))
                    {
                        SampleList list = new SampleList(samples);
                        RenderSampleList(list, name);
                        ok |= (list.Count > 0);
                    }
                }
            }
            catch (System.Exception e)
            {
                MessageBox.Show(e.Message, path, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }

            return ok;
        }

        private bool LoadRemoteSamples(string url)
        {
            bool ok = false;

            try
            {
                // Open an XML reader on the specified URL
                XmlTextReader reader = new XmlTextReader(url);

                // Load the XML into a DataSet
                DataSet ds = new DataSet();
                ds.ReadXml(reader);

                // Sort the default view
                DataTable dt = ds.Tables[0];
                dt.DefaultView.Sort = "Product, Title, Lang";

                // Build an array of distinct product names
                ArrayList names = new ArrayList();
                foreach (DataRowView row in dt.DefaultView)
                {
                    string name = row["Product"].ToString();
                    string platform = row["Platform"].ToString();

                    if (platform.Contains(".NET") && !names.Contains(name))
                        names.Add(name);
                }

                // Render the samples, grouped by product
                foreach (string name in names)
                {
                    dt.DefaultView.RowFilter = "Product = '" + name + "'";
                    SampleList list = new SampleList(dt.DefaultView);
                    RenderSampleList(list, name);
                    ok |= (list.Count > 0);
                }
            }
            catch (System.Exception e)
            {
                MessageBox.Show(e.Message, url, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }

            return ok;
        }

        private void RenderSampleList(SampleList list, string product)
        {
            if (list.Count > 0)
            {
                // Render the container for this product and add it to the document
                RenderObject roProduct = CreateProduct(product);
                Body.Children.Add(roProduct);

                // Add a top-level outline node for this product
                int n = this.Outlines.Add(product, roProduct);
                OutlineNode nodeProduct = this.Outlines[n];

                // Render a top-level toc item for this product, if a toc exists
                if (toc != null)
                {
                    RenderTocItem ti = toc.AddItem(product, roProduct);
                    ti.Style.Font = new Font(baseFont, 10, FontStyle.Bold);
                    ti.Style.Spacing.Top = "10pt";
                }

                // Render each sample for this product
                for (int i = 0; i < list.Count; i++)
                {
                    SampleEntry sample = (SampleEntry)list.GetByIndex(i);
                    RenderObject ro = CreateSample(sample);
                    
                    // Add the sample's RenderObject to the product's container and outline node
                    roProduct.Children.Add(ro);
                    nodeProduct.Children.Add(sample.Name, ro);
                    
                    // Add a second-level toc entry, if a toc exists
                    if (toc != null)
                        toc.AddItem(sample.Name, ro, 2);
                    
                    // Add the sample's RenderObject to the index, if one exists
                    if (indexList != null)
                        AddToIndex(sample.Name, product, ro);
                }
            }
        }

        #region RenderObject creation members
        private RenderObject CreateProduct(string product)
        {
            // Create a RenderArea to use as a container for this
            // product's samples, and to manage page headers/footers
            RenderArea ra = new RenderArea();

			// Render the product name and add it to the RenderArea
            RenderText text = new RenderText(product);
            text.Style.Font = new Font(baseFont, 16, FontStyle.Bold);
            text.Style.Spacing.Bottom = "20pt";
            ra.Children.Add(text);

            // Create a new page layout for this product and set its footer
            PageLayout pl = new PageLayout();
            pl.PageFooter = CreatePageFooter(product);

            if (!firstPage)
            {
                firstPage = true;
                PageNumberingChange numbers = new PageNumberingChange(PageNumberingChangeModeEnum.Set, 1);
                pl.PageFooter.PageNumberingChange = numbers;
            }

            // Force a page break before the RenderArea container
            ra.LayoutChangeBefore = new LayoutChangeNewPage(pl);
            return ra;
        }

        private RenderToc CreateToc()
        {
            RenderArea ra = new RenderArea();

            RenderToc toc = new RenderToc();
            toc.BreakAfter = BreakEnum.Page;
            toc.Style.Font = new Font(baseFont, 10);
            toc.LevelSpacing = "0.25in";

            RenderText text = new RenderText("Table of Contents");
            text.Style.Font = new Font(baseFont, 16, FontStyle.Bold);
            text.Style.Spacing.Bottom = "20pt";
            ra.Children.Add(text);

            ra.Children.Add(toc);
            Body.Children.Add(ra);
            Outlines.Add("Table of Contents", ra);

            // Create a new page layout for this product and set its footer
            PageLayout pl = new PageLayout();
            pl.PageFooter = CreatePageFooter("Table of Contents");
            ra.LayoutChangeBefore = new LayoutChangeNewPage(pl);
            return toc;
        }

        private RenderArea CreateIndex()
        {
            RenderArea indexArea = new RenderArea();
            indexArea.BreakBefore = BreakEnum.Page;
            indexArea.Style.Font = new Font(baseFont, 8);

            RenderText text = new RenderText("Index");
            text.Style.Font = new Font(baseFont, 16, FontStyle.Bold);
            text.Style.Spacing.Bottom = "20pt";
            indexArea.Children.Add(text);

            for (int i = 0; i < indexList.Count; i++)
            {
                string letter = indexList.Keys[i];
                SortedList samples = indexList.Values[i];

                RenderText group = new RenderText(letter);
                group.Style.Font = new Font(baseFont, 10, FontStyle.Bold);
                group.Style.Spacing.Top = "10pt";

                indexArea.Children.Add(group);
                RenderToc index = new RenderToc();
                index.BreakBefore = BreakEnum.None;

                for (int j = 0; j < samples.Count; j++)
                {
                    string entry = samples.GetKey(j).ToString();
                    RenderObject ro = samples.GetByIndex(j) as RenderObject;
                    index.AddItem(entry, ro);
                }
                indexArea.Children.Add(index);
            }

            Body.Children.Add(indexArea);
            Outlines.Add("Index", indexArea);

            // Create a new page layout for this product and set its footer
            PageLayout pl = new PageLayout();
            pl.PageFooter = CreatePageFooter("Index");
            pl.Columns.Add();
            pl.Columns.Add();

            // Force a page break before the RenderArea container
            indexArea.LayoutChangeBefore = new LayoutChangeNewPage(pl);
            return indexArea;
        }

        private RenderObject CreateSample(SampleEntry sample)
		{
			// Create a RenderTable to display the sample as follows:
			// Row 0, Col 0: Nested title table
			// Row 0, Col 1: Links to solutions/projects
			// Row 1: Nested detail/screenshot table
			RenderTable table = new RenderTable(2, 2);

			// Create a nested RenderTable for the sample name and description
			TableCell cellTitle = table.Cells[0, 0];
			cellTitle.RenderObject = CreateTitleTable(sample);

            // Right-align the cell used for project hyperlinks
			TableCell cellLinks = table.Cells[0, 1];
			cellLinks.Style.TextAlignHorz = AlignHorzEnum.Right;
			cellLinks.Style.Padding.Top = "3pt";

			// Create a RenderParagraph to use as a hyperlink container
			RenderParagraph para = new RenderParagraph(cellLinks.Style);
			cellLinks.RenderObject = para;

			// Render an image hyperlink for each solution/project
			foreach (DictionaryEntry e in sample.Projects)
			{
				string language = e.Key.ToString(); // CS or VB
				string target = e.Value.ToString(); // Full path

				// Insert some spaces before the image
				para.Content.AddText("   ");

				// Create a hyperlink target for a specific version of Visual Studio
                string vsprog = VSLauncher.Instance.GetProgram(target);
                C1LinkTargetFile vslink;
                
                if (vsprog != target)
                    vslink = new C1LinkTargetFile(vsprog, "", "\"" + target + "\"", "");
                else
                    vslink = new C1LinkTargetFile(target);

                // Insert the image hyperlink
                ParagraphImage img = para.Content.AddImage(language);
                img.Hyperlink = new C1Hyperlink(vslink, target);
			}

			// Create another nested RenderTable for the detail and screenshot
			TableCell cellDetails = table.Cells[1, 0];
			cellDetails.RenderObject = CreateDetailTable(sample);
            cellDetails.SpanCols = 2;

            // Set the width of the first column
            table.Cols[0].Width = new Unit(75, DimensionEnum.Width);

            // Modify the table's border and spacing and return it
			table.Style.Borders.Top = new LineDef("1pt", Color.LightGray);
			table.Style.Spacing.Bottom = "16pt";
            table.CanSplitVert = true;

            table.Rows[0].PageBreakBehavior = PageBreakBehaviorEnum.PreferredBreak;
            table.Rows[1].PageBreakBehavior = PageBreakBehaviorEnum.NoBreak;
            table.Rows[1].CanSplit = true;
			return table;
		}

        private RenderTable CreateTitleTable(SampleEntry sample)
		{
			// Create a two-row RenderTable for the name and short description
			RenderTable table = new RenderTable(2, 1);

			// Render the sample name in the top cell
			TableCell cellTitle = table.Cells[0, 0];
			cellTitle.Style.Font = new Font(baseFont, 12, FontStyle.Bold);
			cellTitle.Text = sample.Name;

			// Render the short description in the bottom cell
			TableCell cellDesc = table.Cells[1, 0];
			cellDesc.Style.Font = new Font(baseFont, 10);
			cellDesc.Style.Spacing.Top = "2pt";
			cellDesc.Text = sample.Description;

            table.Rows[0].PageBreakBehavior = PageBreakBehaviorEnum.PreferredBreak;
            table.Rows[1].PageBreakBehavior = PageBreakBehaviorEnum.NoBreak;
            return table;
		}

        private RenderTable CreateDetailTable(SampleEntry sample)
		{
			// Create a two-column RenderTable for the detail text and optional image
			RenderTable table = new RenderTable(1, 2);

			// Render the sample's long description in the left cell
			TableCell cellDetails = table.Cells[0, 0];
			FormatTableCell(cellDetails, sample.Details, AlignHorzEnum.Left, AlignVertEnum.Top);
            cellDetails.Style.TextColor = Color.DimGray;
			cellDetails.RenderObject.Style.MinOrphanLines = 3;

			// Render the screenshot image in the right cell, if present
			if (sample.Image != null)
			{
                // Scale and right-align the image
				TableCell cellImage = table.Cells[0, 1];
				RenderImage ri = new RenderImage(sample.Image);
				ri.Style.ImageAlign.BestFit = true;
                ri.Style.ImageAlign.AlignHorz = ImageAlignHorzEnum.Right;
                ri.Width = "100%";
                ri.Height = new Unit(1.75, UnitTypeEnum.Inch);

				// Nest the image in a RenderArea to keep it from being split
				RenderArea ra = new RenderArea();
				ra.Width = "100%";
				ra.Children.Add(ri);
                ra.Style.Padding.Left = "8pt";
				cellImage.RenderObject = ra;
			}
			// Otherwise, the long description spans the entire row
			else
			{
                table.Cols.Delete(1, 1);
                table.Cols[0].Width = new Unit(100, DimensionEnum.Width);
			}

            table.Style.Spacing.Top = "8pt";
            table.Rows[0].PageBreakBehavior = PageBreakBehaviorEnum.NoBreak;
            table.Rows[0].CanSplit = true;
            return table;
		}

        private RenderObject CreatePageHeader()
        {
            // Render the page header as a two-column table with one row
            RenderTable table = new RenderTable(1, 2);
            TableCell left = table.Cells[0, 0];
            TableCell right = table.Cells[0, 1];
            FormatTableCell(left, "C1SampleViewer", AlignHorzEnum.Left, AlignVertEnum.Top);
            FormatTableCell(right, "Created with Preview for .NET 2.0", AlignHorzEnum.Right, AlignVertEnum.Top);
            table.Height = "0.5in";
            table.Rows[0].Style.Borders.Bottom = new LineDef("1pt", Color.Crimson);
            return table;
        }

        private RenderObject CreatePageFooter(string product)
        {
            // Render the page footer as a two-column table with one row
            RenderTable table = new RenderTable(1, 2);
            TableCell left = table.Cells[0, 0];
            TableCell right = table.Cells[0, 1];
            FormatTableCell(left, product, AlignHorzEnum.Left, AlignVertEnum.Bottom);
            FormatTableCell(right, "Page [PageNo] of [PageCount]", AlignHorzEnum.Right, AlignVertEnum.Bottom);
            table.Style.Borders.Top = new LineDef("1pt", Color.Crimson);
            return table;
        }
        #endregion

        private void FormatTableCell(TableCell cell, string text, AlignHorzEnum horz, AlignVertEnum vert)
        {
            cell.Style.TextAlignHorz = horz;
            cell.Style.TextAlignVert = vert;
            cell.Style.Font = new Font(baseFont, 8);
            cell.Style.TextColor = Color.SteelBlue;
            cell.Text = text;
        }
    }
    
    public class SampleList : SortedList
    {
        public SampleList(string path) : base()
        {
            DirectoryInfo info = new DirectoryInfo(path);

            foreach (DirectoryInfo folder in info.GetDirectories())
            {
                foreach (DirectoryInfo sample in folder.GetDirectories())
                {
                    SampleEntry entry = (SampleEntry)this[sample.Name];

                    if (entry == null)
                    {
                        entry = new SampleEntry(sample);

                        if (entry.IsValid)
                            this.Add(sample.Name, entry);
                    }

                    if (entry.IsValid)
                        entry.AddProject(sample);
                }
            }
        }

        public SampleList(DataView view) : base()
        {
            foreach (DataRowView row in view)
            {
                string name = row["Title"].ToString();
                SampleEntry entry = (SampleEntry)this[name];

                if (entry == null)
                {
                    entry = new SampleEntry(row);

                    if (entry.IsValid)
                        this.Add(name, entry);
                }

                if (entry.IsValid)
                    entry.AddProject(row);
            }
        }
    }

    public class SampleEntry
    {
        bool isValid = false;
        string readmePath = null;
        string imagePath = null;
        string name = null;
        string fullName = null;
        string desc = null;
        string details = null;
		SortedList projects = null;
		Image image = null;

        public SampleEntry(DirectoryInfo info)
        {
            name = info.Name;
            fullName = info.FullName;
            readmePath = fullName + @"\readme.txt";
            imagePath = fullName + @"\screenshot.png";
            isValid = File.Exists(readmePath);

            if (isValid)
            {
                string[] lines = File.ReadAllLines(readmePath);

                if (lines.Length > 2)
                {
                    StringBuilder sb = new StringBuilder();

                    for (int i = 2; i < lines.Length; i++)
                    {
                        string s = lines[i];

                        if (desc == null && s.Trim().Length == 0)
                        {
                            sb = sb.Replace("\r\n", " ");
                            desc = sb.ToString();
                            sb.Length = 0;
                            continue;
                        }

                        sb.AppendLine(s.Trim());
                    }

                    sb = sb.Replace("\r\n\r\n", "@@@@");
                    sb = sb.Replace("\r\n", " ");
                    sb = sb.Replace("@@@@", "\r\n\r\n");
                    details = sb.ToString();
                }
            }
        }

        public SampleEntry(DataRowView row)
        {
            name = row["Title"].ToString();
            fullName = row["Path"].ToString();
            desc = row["Description"].ToString();
            details = row["Details"].ToString();
            isValid = true;

            details = details.Replace("\r\n\r\n", "@@@@");
            details = details.Replace("\r\n", " ");
            details = details.Replace("@@@@", "\r\n\r\n");
            details = details.Replace("<p>", "");
            details = details.Replace("</p>", "");

            bool screenshot = System.Convert.ToBoolean(row["Screenshot"]);
            imagePath = screenshot ? fullName + @"/screenshot.png" : "";
        }

        static string[] projectTypes = { "*.sln", "*.csproj", "*.vbproj" };

        public void AddProject(DirectoryInfo info)
        {
            if (projects == null)
                projects = new SortedList();

            for (int i = 0; i < projectTypes.Length; i++)
            {
                FileInfo[] files = info.GetFiles(projectTypes[i]);

                if (files.Length == 1)
                {
                    projects.Add(info.Parent.Name.ToUpper(), files[0].FullName);
                    return;
                }
            }

            projects.Add(info.Parent.Name.ToUpper(), info.FullName);
        }

        public void AddProject(DataRowView row)
        {
            if (projects == null)
                projects = new SortedList();

            string lang = row["Lang"].ToString();
            string url = row["Url"].ToString();

            lang = lang.Replace("C#", "CS");
            projects.Add(lang, url);
        }

        public Image Image
        {
            get
            {
                try
                {
                    if (image == null) // && File.Exists(imagePath))
                    {
                        PictureBox pb = new PictureBox();
                        pb.Load(imagePath);
                        image = pb.Image;
                    }
                }
                catch
                {
                }
                return image;
            }
        }

        public bool IsValid { get { return isValid; } }
        public string Name { get { return name; } }
        public string FullName { get { return fullName; } }
        public string Description { get { return desc; } }
        public string Details { get { return details; } }
        public SortedList Projects { get { return projects; } }
    }
}
